#	jelnet
#	Copyright (C) 2006 Michal Vaner (michal.vaner@kdemail.net)
#
#	This program is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	he Free Software Foundation; either version 2 of the License, or
#	(at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program; if not, write to the Free Software
#	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import globals
from globals import deb, warn
import socket
import sys
import xml.parsers.expat
import base64
import re
import os
import md5

done = False
state = None
hasBind = False
hasSess = False
header = None

def startConnection():
	deb( "Starting connection" )
	for res in socket.getaddrinfo( globals.server, globals.port, socket.AF_UNSPEC, socket.SOCK_STREAM ):
		af, socktype, proto, canonname, sa = res
		deb( "Trying with this data: " + str( af ) + ", " + str( socket ) + ", " + str( proto ) + ", " + str( canonname ) + ", " + str( sa ) )
		try:
			globals.socket = socket.socket( af, socktype, proto )
		except socket.error, msg:
			globals.socket = None
			warn( "Socket not created, " + str( msg ) )
			continue
		try:
			globals.socket.connect( sa )
		except socket.error, msg:
			globals.socket.close()
			globals.socket = None
			warn( "Could not connect to given address, " + str( msg ) )
			continue
		break
	if globals.socket is None:
		sys.stderr.write( "Could not connect to host\n" )
		sys.exit( 1 )
	deb( "Connected" )
	globals.statePrint( "CONN" )
	globals.socket.setblocking( True )

def startSSL():
	deb( "Starting encryption" )
	global finishedInit
	finishedInit = False
	globals.ssl = socket.ssl( globals.socket )
	globals.statePrint( "ENC ON" )
	globals.readDataFunc = globals.readSsl
	globals.sendDataFunc = globals.writeSsl

mechanisms = []
challenge = None

def startSASL():
	deb( "Sending login information" )
	global mechanisms, state
	if "DIGEST-MD5" in mechanisms:
		deb( "Selected digest-md5" )
		globals.sendData( "<auth xmlns='urn:ietf:params:xml:ns:xmpp-sasl' mechanism='DIGEST-MD5'/>" )
		state = "md5-requested"
		global challenge
		challenge = ""
	elif "PLAIN" in mechanisms:
		deb( "Selected plain" )
		globals.sendData( "<auth xmlns='urn:ietf:params:xml:ns:xmpp-sasl' mechanism='PLAIN'>" + base64.encodestring( re.search( "^([^@]+@[^/]+)(/.*)?", globals.jid ).group( 1 ) + "\0" + re.search( "([^@]*)@.*", globals.jid ).group( 1 ) + "\0" + globals.passwd ) + "</auth>" )
		state = "sasl-finished"
	else:
		deb( "No known mechanisms" )
		sys.stderr.write( "No supported SASL mechanism\n" )
		sys.exit( 3 )
	mechanisms = None


def startElem( name, attrs ):
	global state, hasSess, done
	deb( "Starting element " + name + " " + str( attrs ) )
	if name == "http://etherx.jabber.org/streams:features":
		state = "features"
		deb( "Starting to decode features" )
	elif( name == "urn:ietf:params:xml:ns:xmpp-tls:starttls" ) and ( ( state == "features" ) or ( state == "sasl-avail" ) ):
		state = "pre-tls"
		deb( "Negotiating encryption" )
		globals.sendData( "<starttls xmlns='urn:ietf:params:xml:ns:xmpp-tls'/>" )
	elif( name == "urn:ietf:params:xml:ns:xmpp-tls:proceed" ) and ( state == "pre-tls" ):
		done = True
		startSSL()
	elif( name == "urn:ietf:params:xml:ns:xmpp-sasl:mechanisms" ) and ( state == "features" ):
		state = "sasl-avail"
		global mechanisms
		mechanisms = []
	elif( name == "urn:ietf:params:xml:ns:xmpp-sasl:mechanism" ) and ( state == "sasl-avail" ):
		mechanisms.append( "" )
	elif( name == "urn:ietf:params:xml:ns:xmpp-sasl:challenge" ) and ( state == "md5-requested" ):
		global challenge
		challenge = ""
		state = "challenge-md5"
	elif( name == "urn:ietf:params:xml:ns:xmpp-bind:bind" ):
		global hasBind
		hasBind = True
	elif( name == "urn:ietf:params:xml:ns:xmpp-session:session" ):
		hasSess = True
	elif( name == "jabber:client:iq" ) and ( attrs[ "id" ] == "bind" ):
		if attrs[ "type" ] == "result":
			deb( "Resource binded" )
			state = "bind-data"
			if hasSess:
				doSess()
		else:
			sys.stderr.write( "Could not bind address" )
			print "-h,--header\t\t\tPrint the stream header for xml parser."
			sys.exit( 4 )
	elif( name == "jabber:client:iq" ) and ( attrs[ "id" ] == "session" ):
		if attrs[ "type" ] == "result":
			deb( "Session established" )
			globals.statePrint( "SESS DONE" )
			done = True
		else:
			sys.stderr.write( "Could not create session" )
			sys.exit( 5 )

def doBind():
	deb( "Binding resource" )
	resource = None
	mo = re.search( "^[^@]+@[^/]+/(.*)$", globals.jid )#try find the resource
	if mo:
		resource = mo.group( 1 )
	req = "<iq type='set' id='bind'><bind xmlns='urn:ietf:params:xml:ns:xmpp-bind'"
	if resource:
		req = req + "><resource>" + resource + "</resource></bind></iq>"
	else:
		req = req + "/></iq>"
	global state
	state = "bind"
	globals.sendData( req )

def doSess():
	deb( "Session start" )
	globals.sendData( "<iq type='set' id='session'><session xmlns='urn:ietf:params:xml:ns:xmpp-session'/></iq>" )

def endElem( name ):
	global state, done
	deb( "Ending element " + name )
	if( name == "http://etherx.jabber.org/streams:features" ) and ( state == "features" ):
		state = None
		deb( "Finished features" )
		global hasSess, hasBind
		if hasBind and globals.doBind:
			doBind()
		elif hasSess and globals.doSess:
			doSess()
		else:
			done = True
	elif( name == "http://etherx.jabber.org/streams:features" ) and ( state == "sasl-avail" ):
		state = "sasl"
		startSASL()
	elif( name == "urn:ietf:params:xml:ns:xmpp-sasl:challenge" ) and ( state == "challenge-md5" ):
		state = "md5-requested"
		global challenge
		chall = base64.decodestring( challenge )
		mo = re.search( "nonce=['\"]([^'\"]*)['\"]", chall )
		if mo:
			nonce = mo.group( 1 )
			cnonce = base64.encodestring( open( "/dev/random", "r" ).read( 10 ) )
			realm = globals.extractServer( globals.jid )
			nc = "00000001"
			uri = "xmpp/" + realm
			qop = "auth"

			node = re.search( "([^@]*)@.*", globals.jid ).group( 1 )
			x = node + ":" + realm + ":" + globals.passwd
			y = md5.new( x ).digest()
			tmp = ":" + nonce + ":" + cnonce + ":jelnet"
			y = y + tmp
			a2 = "AUTHENTICATE:" + uri
			ha1 = md5.new( y ).hexdigest()
			ha2 = md5.new( a2 ).hexdigest()
			kd = ha1 + ":" + nonce + ":" + nc + ":" + cnonce + ":" + qop + ":" + ha2
			z = md5.new( kd ).hexdigest()
			result = "username=\"" + node + "\",realm=\"" + realm + "\",nonce=\"" + nonce + "\",cnonce=\"" + cnonce + "\",nc=\"" + nc + "\",serv-type=\"xmpp\",host=\"" + realm + "\",digest-uri=\"" + uri + "\",qop=\"" + qop + "\",response=\"" + z + "\",charset=\"utf-8\",authzid=\"jelnet\""
			globals.sendData( "<response xmlns='urn:ietf:params:xml:ns:xmpp-sasl'>" + base64.encodestring( result ) + "</response>" )
		else:
			globals.sendData( "<response xmlns='urn:ietf:params:xml:ns:xmpp-sasl'/>" )
			state = "sasl-finished"
	elif( name == "urn:ietf:params:xml:ns:xmpp-sasl:success" ) and ( state == "sasl-finished" ):
		globals.statePrint( "LOGGED" )
		deb( "Logged in" )
		done = True
		global finishedInit
		finishedInit = not globals.doBind
	elif( name == "urn:ietf:params:xml:ns:xmpp-sasl:failure" ):
		globals.statePrint( "BAD AUTH" )
		deb( "Bad auth" )
		sys.exit( 3 )

def cdata( data ):
	deb( "CData: " + data )
	global state
	if state == "sasl-avail":
		global mechanisms
		if mechanisms != []:
			mechanisms[ len( mechanisms ) - 1 ] = mechanisms[ len( mechanisms ) - 1 ] + data.strip()
	elif state == "challenge-md5":
		global challenge
		challenge = challenge + data.strip()
	elif state == "bind-data":
		if data.strip() != "":
			globals.statePrint( "JID " + data.strip() )

def initStartParser():
	global parser
	parser = xml.parsers.expat.ParserCreate( "UTF-8", ":" )
	parser.StartElementHandler = startElem
	parser.EndElementHandler = endElem
	parser.CharacterDataHandler = cdata
	parser.returns_unicode = False

def parseData( data ):
	global done
	done = False
	global parser
	parser.Parse( data, False )
	if done:
		parser = None
		return False
	return True

def startAndProcessStream():
	deb( "Starting the XML stream" )
	globals.sendData( "<?xml version='1.0'?><stream:stream to='" + globals.extractServer( globals.jid ) + "' xmlns='jabber:client' xmlns:stream='http://etherx.jabber.org/streams' version='1.0'>" )
	initStartParser()
	global finishedInit
	finishedInit = True
	global hasSess, hasBind
	hasBind = hasSess = False
	global header
	header = ""
	data = globals.readMore()
	while( parseData( data ) ):
		header = header + data
		data = globals.readMore()
	return not finishedInit

def login():
	deb( "Starting login" )
	if( globals.oldSSL ):
		startSSL()
	while( startAndProcessStream() ):
		pass

def cleanLogin():
	global parser, mechanisms
	parser = None
	mechanisms = None
	if globals.stateInfo:
		global header
		header = re.sub( "(id=['\"]([^'\"]*)[\"'])", "", header )
		sys.stderr.write( "INF: HDR " + re.search( "^((<\\?xml[^>]*>)?\\s*<[^>]*>)", header ).group( 1 ) )
